/*
 * Copyright (C) 2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.jaguar.lifecycle;

import static org.eiichiro.jaguar.Jaguar.*;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.Collection;

import org.eiichiro.jaguar.Container;
import org.eiichiro.jaguar.Descriptor;
import org.eiichiro.jaguar.inject.Inject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Lifecycle event that fires on a component.
 * If lifecycle event fires, the component methods listening to the lifecycle 
 * (methods annotated with the lifecycle) are called back.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public abstract class Event<L extends Annotation> {

	private Logger logger = LoggerFactory.getLogger(getClass());
	
	protected final Class<L> lifecycle;
	
	protected Object arg;
	
	protected Object[] components = {};
	
	@Inject private Container container;
	
	/** Constructs a new {@code Event}. */
	public Event() {
		this(null);
	}
	
	/** Constructs a new {@code Event}. */
	@SuppressWarnings("unchecked")
	public Event(Object arg) {
		Type type = getClass().getGenericSuperclass();
		
		for (Class<?> superclass = getClass().getSuperclass(); 
				!superclass.equals(Event.class); superclass = superclass.getSuperclass()) {
			if (type instanceof ParameterizedType) {
				type = ((Class<?>) ((ParameterizedType) type).getRawType()).getGenericSuperclass();
			} else {
				type = ((Class<?>) type).getGenericSuperclass();
			}
		}
		
		Type lifecycle = ((ParameterizedType) type).getActualTypeArguments()[0];
		
		if (lifecycle instanceof Class<?> && ((Class<?>) lifecycle).isAnnotationPresent(Lifecycle.class)) {
			this.lifecycle = (Class<L>) lifecycle;
		} else {
			throw new IllegalArgumentException(
					"Type parameter [L] of [Event<L>] must be a lifecycle annotation type, "
							+ "not be [Event<" + lifecycle + ">]");
		}
		
		this.arg = arg;
		assemble(this);
	}
	
	/**
	 * Sets components on which this event fires.
	 * 
	 * @param components The components on which this event fires.
	 * @return This {@code Event} instance.
	 */
	public Event<L> on(Object... components) {
		this.components = components;
		return this;
	}
	
	/**
	 * Sets components on which this event fires.
	 * 
	 * @param components The components on which this event fires.
	 * @return This {@code Event} instance.
	 */
	public Event<L> on(Collection<Object> components) {
		this.components = components.toArray();
		return this;
	}
	
	/**
	 * Calls back the methods listening to the lifecycle on the specified 
	 * components.
	 * If an exception has occurred, this method catches it and raises 
	 * {@link Failed} event on the component.
	 */
	public final void fire() {
		for (Object component : components) {
			Descriptor<?> descriptor = container.descriptors().get(component);
			logger.debug("Event [" + lifecycle.getSimpleName()
					+ "] fires on component [" + descriptor + "]");
			
			for (Method method : descriptor.lifecycles().get(lifecycle)) {
				if (!method.isAccessible()) {
					method.setAccessible(true);
				}
				
				try {
					if (method.getParameterTypes().length == 0) {
						method.invoke(component);
					} else {
						method.invoke(component, arg);
					}
					
					continue;
				} catch (InvocationTargetException e) {
					logger.warn("Exception from [" + lifecycle.getSimpleName()
							+ " (" + method + ")] event listener on component ["
							+ descriptor + "]", e.getTargetException());
					if (!lifecycle.equals(Failed.class)) {
						new Event<Failed>(e.getTargetException()) {}.on(component).fire();
					}
					
				} catch (Exception e) {
					logger.warn("Cannot fire [" + lifecycle.getSimpleName()
							+ " (" + method + ")] event on component ["
							+ descriptor + "]", e);
				}
			}
		}
	}
	
}
